#include "Dict.h"

#include <iostream>

using namespace std;

namespace dictionary {

struct Dict::Node
{
    string key;
    unsigned value;
    Node* left;
    Node* right;

    static void deep_delete(Node*);
    static Node* deep_clone(const Node*);
};

void Dict::Node::deep_delete(Node* node)
{
    if (node) {
        deep_delete(node->left);
        deep_delete(node->right);
        delete node;
    }
}

Dict::Node* Dict::Node::deep_clone(const Node* node)
{
    if (node == nullptr) return nullptr;

    Node* result = new Node;

    result->key   = node->key;
    result->value = node->value;
    result->left  = deep_clone(node->left);
    result->right = deep_clone(node->right);

    return result;
}

Dict::Dict() : size_(0), root_(nullptr)
{ }

Dict::Dict(const Dict& that)
        : size_(that.size_),
          root_(Node::deep_clone(that.root_))
{ }

Dict& Dict::operator=(const Dict& that)
{
    Node::deep_delete(root_);
    size_ = that.size_;
    root_ = Node::deep_clone(that.root_);

    return *this;
}

Dict::~Dict()
{
    Node::deep_delete(root_);
}

bool Dict::empty() const {
    return root_ == nullptr;
}

size_t Dict::size() const {
    return size_;
}

void Dict::clear()
{
    Node::deep_delete(root_);
    root_ = nullptr;
    size_ = 0;
}

bool Dict::member(const string& key) const
{
    return find_subtree(root_, key) != nullptr;
}

unsigned Dict::lookup(const string& key) const
{
    Node* node = find_subtree(root_, key);
    if (node == nullptr)
        throw Not_found(key);
    else return node->value;
}

unsigned& Dict::lookup(const std::string& key)
{
    Node* node = find_subtree(root_, key);
    if (node == nullptr)
        throw Not_found(key);
    else return node->value;
}

void Dict::insert(const string& key, unsigned value)
{
    Node* & node = find_subtree_mut(root_, key);
    if (node == nullptr) {
        node = new Node;
        node->key = key;
        node->left = node->right = nullptr;
        ++size_;
    }
    node->value = value;
}

unsigned& Dict::operator[](const std::string& key)
{
    Node* & node = find_subtree_mut(root_, key);
    if (node == nullptr) {
        node = new Node;
        node->key = key;
        node->value = 0;
        node->left = node->right = nullptr;
        ++size_;
    }

    return node->value;
}

Dict::Node* Dict::find_subtree(Dict::Node* link, const string& key)
{
    while (link != nullptr && link->key != key)
        link = (key < link->key) ? link->left : link->right;

    return link;
}

Dict::Node*& Dict::find_subtree_mut(Dict::Node*& link, const string& key)
{
    if (link == nullptr || link->key == key)
        return link;

    Dict::Node* curr = link;
    for (;;) {
        if (key < curr->key) {
            if (curr->left == nullptr || key == curr->left->key)
                return curr->left;
            else
                curr = curr->left;
        } else {
            if (curr->right == nullptr || key == curr->right->key)
                return curr->right;
            else
                curr = curr->right;
        }
    }
}

void Dict::get_keys_helper(Dict::Node* node, vector<string>& out)
{
    if (node != nullptr) {
        get_keys_helper(node->left, out);
        out.push_back(node->key);
        get_keys_helper(node->right, out);
    }
}

vector<string> Dict::get_keys() const
{
    vector<string> result;
    get_keys_helper(root_, result);
    return result;
}

}
